import pytest

from threescale_api.resources import ApplicationPlan


@pytest.fixture()
def pricing_rules(metric, application_plan: ApplicationPlan):
    params = dict(min=10, max=100, cost_per_unit=20)
    application_plan.pricing_rules(metric).create(params)
    return application_plan.pricing_rules(metric).list()


def test_create_pricing_rule(pricing_rules):
    assert pricing_rules is not None
    rule = pricing_rules[0]
    assert rule['max'] == 100
    assert rule['min'] == 10
    assert rule['cost_per_unit'] == '20.0'



class ThreeScaleApiError(Exception):
    def __init__(self, message, *args):
        self.message = message
        super(ThreeScaleApiError, self).__init__(message, *args)


class ApiClientError(ThreeScaleApiError):
    def __init__(self, code, body, message: str = None):
        self.code = code
        self.body = body
        self._message = message
        msg = f"Response({self.code}): {body}"
        if message:
            msg += f"; {message}"
        super(ApiClientError, self).__init__(msg)



import logging
from typing import Dict, List, Optional, TYPE_CHECKING, Union, Any, Iterator

import collections.abc

import requests

from threescale_api import utils

if TYPE_CHECKING:
    from threescale_api.client import ThreeScaleClient, RestApiClient

log = logging.getLogger(__name__)


class DefaultClient(collections.abc.Mapping):
    def __init__(self, parent=None, instance_klass=None,
                 entity_name: str = None, entity_collection: str = None):
        """Creates instance of the default client
        Args:
            parent: Parent resource or client
            instance_klass: Which class should be used to instantiate the resource
            entity_name(str): Entity name - required for extraction
            entity_collection(str): Collection name - required for extraction
        """
        self._parent = parent
        self._instance_klass = instance_klass
        self._entity_name = entity_name
        if entity_collection is None and entity_name is not None:
            entity_collection = f'{entity_name}s'
        self._entity_collection = entity_collection

    @property
    def url(self) -> str:
        """Default url for the resources collection
        Returns(str): URL
        """
        return self.threescale_client.admin_api_url

    @property
    def threescale_client(self) -> 'ThreeScaleClient':
        """Gets instance of the 3scale default client
        Returns(TheeScaleClient): 3scale client

        """
        return self.parent.threescale_client

    @property
    def parent(self) -> 'DefaultResource':
        """ Instance of the parent resource
        Returns(DefaultResource): Parent of the client is an subclass of the default resource

        """
        return self._parent

    @property
    def rest(self) -> 'RestApiClient':
        """Rest API client for the 3scale instance
        Returns(RestApiClient):

        """
        return self.threescale_client.rest

    def list(self, **kwargs) -> List['DefaultResource']:
        """List all entities
        Args:
            **kwargs: Optional parameters
        Returns(List['DefaultResource]): List of resources
        """
        log.info(self._log_message("[LIST] List", args=kwargs))
        instance = self._list(**kwargs)
        return instance

    def create(self, params: dict = None, **kwargs) -> 'DefaultResource':
        """Create a new instance
        Args:
            params: Parameters required to create new instance
            **kwargs: Optional parameters

        Returns:

        """
        log.info(self._log_message("[CREATE] Create new ", body=params, args=kwargs))
        url = self._entity_url()
        response = self.rest.post(url=url, json=params, **kwargs)
        instance = self._create_instance(response=response)
        return instance

    def delete(self, entity_id: int = None, **kwargs) -> bool:
        """Delete resource
        Args:
            entity_id(int): Entity id
            **kwargs: Optional args

        Returns(bool): True if the resource has been successfully deleted
        """
        log.info(self._log_message("[DELETE] Delete ", entity_id=entity_id, args=kwargs))
        url = self._entity_url(entity_id=entity_id)
        response = self.rest.delete(url=url, **kwargs)
        return response.ok

    def exists(self, entity_id=None, **kwargs) -> bool:
        """Check whether the resource exists
        Args:
            entity_id(int): Entity id
            **kwargs: Optional args

        Returns(bool): True if the resource exists
        """
        log.info(self._log_message("[EXIST] Resource exist ", entity_id=entity_id, args=kwargs))
        url = self._entity_url(entity_id=entity_id)
        response = self.rest.get(url=url, throws=False, **kwargs)
        return response.ok

    def update(self, entity_id=None, params: dict = None, **kwargs) -> 'DefaultResource':
        """Update resource
        Args:
            entity_id(int): Entity id
            params(dict): Params to be updated
            **kwargs: Optional args

        Returns(DefaultResource): Resource instance
        """
        log.info(self._log_message("[UPDATE] Update ", body=params,
                                   entity_id=entity_id, args=kwargs))
        url = self._entity_url(entity_id=entity_id)
        response = self.rest.put(url=url, json=params, **kwargs)
        instance = self._create_instance(response=response)
        return instance

    def fetch(self, entity_id: int = None, **kwargs) -> dict:
        """Fetch the entity dictionary
        Args:
            entity_id(int): Entity id
            **kwargs: Optional args

        Returns(dict): Resource dict from the 3scale
        """
        log.debug(self._log_message("[FETCH] Fetch ", entity_id=entity_id, args=kwargs))
        url = self._entity_url(entity_id=entity_id)
        response = self.rest.get(url=url, **kwargs)
        return utils.extract_response(response=response, entity=self._entity_name)

    def __getitem__(self, selector: Union[int, 'str']) -> 'DefaultResource':
        """Gets the item
        Args:
            selector(Union[int, 'str']): Selector whether id or string
        Returns(DefaultResource): Resource instance
        """
        if isinstance(selector, int):
            return self.read(selector)
        return self.read_by_name(selector)

    def __len__(self) -> int:
        return len(self._list())

    def __iter__(self) -> Iterator['CRUDResource']:
        return next(iter(self._list()))

    def read(self, entity_id: int = None) -> 'DefaultResource':
        """Read the instance, read will just create empty resource and lazyloads only if needed
        Args:
            entity_id(int): Entity id
        Returns(DefaultResource): Default resource
        """
        log.debug(self._log_message("[READ] Read ", entity_id=entity_id))
        return self._instance_klass(client=self, entity_id=entity_id)

    def read_by_name(self, name: str, **kwargs) -> 'DefaultResource':
        """Read resource by name
        Args:
            name: Name of the resource (either system name, name, org_name ...)
            **kwargs:

        Returns:

        """
        for item in self._list(**kwargs):
            if item.entity_name and item.entity_name == name:
                return item

    def select(self, predicate, **kwargs) -> List['DefaultResource']:
        """Select resource s based on the predicate
        Args:
            predicate: Predicate
            **kwargs: Optional args
        Returns: List of resources
        """
        return [item for item in self._list(**kwargs) if predicate(item)]

    def select_by(self, **params) -> List['DefaultResource']:
        """Select by params - logical and
        Args:
            **params: params used for selection
        Returns: List of resources
        """
        log.debug(f"[SELECT] By params: {params}")

        def predicate(item):
            for (key, val) in params.items():
                if item[key] != val:
                    return False
            return True

        return self.select(predicate=predicate)

    def read_by(self, **params) -> 'DefaultResource':
        """Read by params - it will return just one instance of the resource
        Args:
            **params: params used for selection
        Returns(DefaultResource): Resource instance
        """
        result = self.select_by(**params)
        return result[0] if result else None

    def _log_message(self, message, entity_id=None, body=None, args=None) -> str:
        msg = f"{message} {self._instance_klass.__name__}"
        if entity_id:
            msg += f"({entity_id}))"
        if body:
            msg += f" {body}"
        if args:
            msg += f" args={args}"
        return msg

    def _list(self, **kwargs) -> List['DefaultResource']:
        """Internal list implementation used in list or `select` methods
        Args:
            **kwargs: Optional parameters

        Returns(List['DefaultResource']):

        """
        url = self._entity_url()
        response = self.rest.get(url=url, **kwargs)
        instance = self._create_instance(response=response, collection=True)
        return instance

    def _entity_url(self, entity_id=None) -> str:
        if not entity_id:
            return self.url
        return self.url + '/' + str(entity_id)

    def _create_instance(self, response: requests.Response, klass=None, collection: bool = False):
        klass = klass or self._instance_klass
        extracted = self._extract_resource(response, collection)
        instance = self._instantiate(extracted=extracted, klass=klass)
        log.debug(f"[INSTANCE] Created instance: {instance}")
        return instance

    def _extract_resource(self, response, collection) -> Union[List, Dict]:
        extract_params = dict(response=response, entity=self._entity_name)
        if collection:
            extract_params['collection'] = self._entity_collection
        extracted = utils.extract_response(**extract_params)
        return extracted

    def _instantiate(self, extracted, klass):
        if isinstance(extracted, list):
            instance = [self.__make_instance(item, klass) for item in extracted]
            return instance
        return self.__make_instance(extracted, klass)

    def __make_instance(self, extracted: dict, klass):
        instance = klass(client=self, entity=extracted) if klass else extracted
        return instance


class DefaultResource(collections.abc.MutableMapping):
    def __init__(self, client: DefaultClient = None, entity_id: int = None, entity_name: str = None,
                 entity: dict = None):
        """Create instance of the resource
        Args:
            client: Client instance of the resource
            entity_id(int): Entity id
            entity_name(str): Entity name field (system_name or name ...)
            entity(dict): Entity instance
        """
        self._entity_id = entity_id or entity.get('id')
        self._entity = entity
        self._client = client
        self._entity_name = entity_name

    @property
    def threescale_client(self) -> 'ThreeScaleClient':
        return self.client.threescale_client

    @property
    def parent(self) -> 'DefaultResource':
        return self.client.parent

    @property
    def entity_name(self) -> Optional[str]:
        return self[self._entity_name]

    @property
    def url(self) -> str:
        return self.client.url + f"/{self.entity_id}"

    @property
    def entity(self) -> dict:
        self._lazy_load()
        return self._entity

    @property
    def client(self) -> DefaultClient:
        return self._client

    @property
    def entity_id(self) -> int:
        return self._entity_id or self._entity.get('id')

    def __getitem__(self, item: str):
        return self.entity.get(item)

    def __setitem__(self, key: str, value):
        self.set(key, value)

    def __delitem__(self, key: str):
        del self.entity[key]

    def __len__(self) -> int:
        return len(self.entity)

    def __iter__(self) -> Iterator:
        return iter(self.entity)

    def __str__(self) -> str:
        return self.__class__.__name__ + f"({self.entity_id}): " + str(self.entity)

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other) -> bool:
        return (
            self.__class__ == other.__class__ and
            self.entity_name == other.entity_name and
            self.entity_id == other.entity_id
        )

    def get(self, item):
        return self.entity.get(item)

    def set(self, item: str, value: Any):
        self.entity[item] = value

    def _lazy_load(self, **kwargs) -> 'DefaultResource':
        if not self._entity:
            # Lazy load the entity
            self._entity = self.fetch(**kwargs)
        return self

    def read(self, **kwargs) -> 'DefaultResource':
        self._invalidate()
        self._lazy_load(**kwargs)
        return self

    def fetch(self, **kwargs) -> dict:
        return self.client.fetch(self.entity_id, **kwargs)

    def exists(self, **kwargs) -> bool:
        return self.client.exists(entity_id=self.entity_id, **kwargs)

    def delete(self, **kwargs):
        self.client.delete(entity_id=self.entity_id, **kwargs)

    def update(self, params: dict = None, **kwargs) -> 'DefaultResource':
        new_params = {**self.entity}
        if params:
            new_params.update(params)
        new_entity = self.client.update(entity_id=self.entity_id, params=new_params, **kwargs)
        self._entity = new_entity.entity
        return self

    def _invalidate(self):
        self._entity = None


class DefaultPlanClient(DefaultClient):
    def set_default(self, entity_id: int, **kwargs) -> 'DefaultPlanResource':
        """Sets default plan for the entity
        Args:
            entity_id: Entity id
            **kwargs: Optional args
        Returns(DefaultPlanResource):
        """
        log.info(self._log_message("[PLAN] Set default ", entity_id=entity_id, args=kwargs))
        url = self._entity_url(entity_id) + '/default'
        response = self.rest.put(url=url, **kwargs)
        instance = self._create_instance(response=response)
        return instance

    def get_default(self, **kwargs) -> Optional['DefaultResource']:
        """Get default plan if set
        Args:
            **kwargs: Optional arguments
        Returns(DefaultResource): Resource instance
        """
        default = self.select(lambda x: x.is_default, **kwargs)
        if default:
            return default[0]
        return None


class DefaultPlanResource(DefaultResource):
    def __init__(self, entity_name='system_name', **kwargs):
        super().__init__(entity_name=entity_name, **kwargs)

    def set_default(self, **kwargs) -> 'DefaultStateResource':
        """Set the plan default
        Args:
            **kwargs: Optional args
        Returns(DefaultStateResource): State resource instance
        """
        return self.client.set_default(entity_id=self.entity_id, **kwargs)

    @property
    def is_default(self) -> bool:
        return self['default'] is True


class DefaultStateClient(DefaultClient):
    def set_state(self, entity_id, state: str, **kwargs):
        """Sets the state for the resource
        Args:
            entity_id(int): Entity id
            state(str): Which state
            **kwargs: Optional args

        Returns(DefaultStateResource): State resource instance
        """
        log.info(self._log_message("[STATE] Set state ", body=f"[{state}]", args=kwargs))
        url = self._entity_url(entity_id) + '/' + state
        response = self.rest.put(url=url, **kwargs)
        instance = self._create_instance(response=response)
        return instance


class DefaultStateResource(DefaultResource):
    def set_state(self, state: str, **kwargs) -> 'DefaultStateResource':
        """Sets the state for the resource
        Args:
            state(str): Which state
            **kwargs: Optional args

        Returns(DefaultStateResource): State resource instance
        """
        return self.client.set_state(entity_id=self.entity_id, state=state, **kwargs)


class DefaultUserResource(DefaultStateResource):
    def __init__(self, entity_name='username', **kwargs):
        super().__init__(entity_name=entity_name, **kwargs)

    def suspend(self, **kwargs) -> 'DefaultUserResource':
        """Suspends the user
        Args:
            **kwargs: Optional arguments
        Returns(DefaultUserResource): User instance

        """
        return self.set_state(state='suspend', **kwargs)

    def resume(self, **kwargs):
        """Resumes the user
        Args:
            **kwargs: Optional arguments
        Returns(DefaultUserResource): User instance

        """
        return self.set_state(state='resume', **kwargs)

    def activate(self, **kwargs):
        """Activates the user
        Args:
            **kwargs: Optional arguments
        Returns(DefaultUserResource): User instance

        """
        return self.set_state(state='activate', **kwargs)

    def set_as_admin(self, **kwargs):
        """Promotes the user to admin
        Args:
            **kwargs: Optional arguments
        Returns(DefaultUserResource): User instance

        """
        return self.set_state(state='set_as_admin', **kwargs)

    def set_as_member(self, **kwargs):
        """Demotes the user to s member
        Args:
            **kwargs: Optional arguments
        Returns(DefaultUserResource): User instance

        """
        return self.set_state(state='set_as_member', **kwargs)



import pytest

from threescale_api.errors import ApiClientError

from tests.integration import asserts


def test_should_create_method(method, method_params):
    asserts.assert_resource(method)
    asserts.assert_resource_params(method, method_params)


def test_should_not_create_method_for_custom_metric(metric, method_params):
    resource = metric.methods.create(params=method_params, throws=False)
    asserts.assert_errors_contains(resource, ['parent_id'])


def test_should_friendly_name_be_required(hits_metric):
    resource = hits_metric.methods.create(params={}, throws=False)
    asserts.assert_errors_contains(resource, ['friendly_name'])


def test_should_raise_api_exception(hits_metric):
    with pytest.raises(ApiClientError):
        hits_metric.methods.create(params={})


def test_should_read_method(method, method_params):
    resource = method.read()
    asserts.assert_resource(resource)
    asserts.assert_resource_params(resource, method_params)


def test_should_update_method(method, updated_method_params):
    resource = method.update(params=updated_method_params)
    asserts.assert_resource(resource)
    asserts.assert_resource_params(resource, updated_method_params)


def test_should_delete_method(hits_metric, updated_method_params):
    resource = hits_metric.methods.create(params=updated_method_params)
    assert resource.exists()
    resource.delete()
    assert not resource.exists()


def test_should_list_methods(hits_metric):
    resources = hits_metric.methods.list()
    assert len(resources) == 1



# pylint: disable=R0903
"""Implementation of custom 3scale specific authentication method(s) for requests (api clients"""

import requests
import requests.auth


class BaseClientAuth(requests.auth.AuthBase):
    """Abstract class for authentication of api client"""

    def __init__(self, app, location=None):
        self.app = app
        self.location = location
        if location is None:
            self.location = app.service.proxy.list().entity["credentials_location"]

    def __call__(self, request):
        credentials = self.credentials

        if self.location == "authorization":
            credentials = credentials.values()
            auth = requests.auth.HTTPBasicAuth(*credentials)
            return auth(request)

        if self.location == "headers":
            request.prepare_headers(credentials)
        elif self.location == "query":
            request.prepare_url(request.url, credentials)
        else:
            raise ValueError("Unknown credentials location '%s'" % self.location)

        return request


class UserKeyAuth(BaseClientAuth):
    """Provides user_key authentication for api client calls"""

    def __init__(self, app, location=None):
        super(UserKeyAuth, self).__init__(app, location)
        self.credentials = {
            self.app.service.proxy.list()["auth_user_key"]: self.app["user_key"]
        }

    def __call__(self, request):
        if self.location == "authorization":
            auth = requests.auth.HTTPBasicAuth(next(iter(self.credentials.values())), "")
            return auth(request)
        return super().__call__(request)


class AppIdKeyAuth(BaseClientAuth):
    """Provides app_id/app_key pair based authentication for api client calls"""

    def __init__(self, app, location=None):
        super(AppIdKeyAuth, self).__init__(app, location)
        self.credentials = {
            "app_id": self.app["application_id"],
            "app_key": self.app.keys.list()["keys"][0]["key"]["value"]
        }

    def __call__(self, request):
        if self.location == "authorization":
            credentials = self.credentials
            auth = requests.auth.HTTPBasicAuth(
                credentials["app_id"], credentials["app_key"])
            return auth(request)
        return super().__call__(request)



import pytest
import backoff

from threescale_api.errors import ApiClientError

from tests.integration import asserts


def test_should_create_metric(metric, metric_params):
    asserts.assert_resource(metric)
    asserts.assert_resource_params(metric, metric_params)


def test_should_fields_be_required(service):
    resource = service.metrics.create(params={}, throws=False)
    asserts.assert_errors_contains(resource, ['friendly_name', 'unit'])


def test_should_system_name_be_invalid(service, metric_params):
    metric_params['system_name'] = 'invalid name whitespaces'
    resource = service.metrics.create(params=metric_params, throws=False)
    asserts.assert_errors_contains(resource, ['system_name'])


def test_should_raise_exception(service):
    with pytest.raises(ApiClientError):
        service.metrics.create(params={})


def test_should_read_metric(metric, metric_params):
    resource = metric.read()
    asserts.assert_resource(resource)
    asserts.assert_resource_params(resource, metric_params)


def test_should_update_metric(metric, updated_metric_params):
    resource = metric.update(params=updated_metric_params)
    asserts.assert_resource(resource)
    asserts.assert_resource_params(resource, updated_metric_params)


def test_should_delete_metric(service, updated_metric_params):
    resource = service.metrics.create(params=updated_metric_params)
    assert resource.exists()
    resource.delete()
    assert not resource.exists()


def test_should_list_metrics(service):
    resources = service.metrics.list()
    assert len(resources) > 1


def test_should_apicast_return_403_when_metric_is_disabled(
        service, metric_params, create_mapping_rule,
        account, ssl_verify, backend_usage):
    """Metric is disabled when its limit is set to 0."""

    proxy = service.proxy.list()
    plan = service.app_plans.create(params=dict(name='metrics-disabled'))
    application_params = dict(name='metrics-disabled', plan_id=plan['id'],
                              description='metric disabled')
    app = account.applications.create(params=application_params)

    metric = service.metrics.create(params=metric_params)
    plan.limits(metric).create(params=dict(period='month', value=0))

    rules = proxy.mapping_rules.list()
    for rule in rules:
        rule.delete()
    rule = create_mapping_rule(metric, 'GET', '/foo/bah/')

    update_proxy_endpoint(service, backend_usage)

    # params = get_user_key_from_application(app, proxy)
    client = app.api_client(verify=ssl_verify)
    response = make_request(client, rule['pattern'])
    assert response.status_code == 403


@backoff.on_predicate(backoff.expo, lambda resp: resp.status_code == 200,
                      max_tries=8)
def make_request(client, path):
    return client.get(path=path)


def get_user_key_from_application(app, proxy):
    user_key = app['user_key']
    user_key_param = proxy['auth_user_key']
    return {user_key_param: user_key}


def update_proxy_endpoint(service, backend_usage):
    """Update service proxy."""
    path = backend_usage['path']
    backend_usage['path'] = '/moloko'
    backend_usage.update()
    backend_usage['path'] = path
    backend_usage.update()
    proxy = service.proxy.list().configs.list(env='sandbox').proxy
    proxy.deploy()
    proxy_tmp = service.proxy.list().configs.list(env='sandbox')
    version = proxy_tmp.entity['proxy_configs'][-1]['proxy_config']['version']
    proxy.promote(version=version)


def test_should_apicast_return_429_when_limits_exceeded(
        service, application_plan, create_mapping_rule,
        apicast_http_client, backend_usage):
    metric_params = dict(system_name='limits_exceeded', unit='count',
                         friendly_name='Limits Exceeded')
    metric = service.metrics.create(params=metric_params)
    application_plan.limits(metric).create(params=dict(period='day', value=1))

    rule = create_mapping_rule(metric, 'GET', '/limits/exceeded/')

    update_proxy_endpoint(service, backend_usage)

    response = apicast_http_client.get(path=rule['pattern'])
    while response.status_code == 200:
        response = apicast_http_client.get(path=rule['pattern'])

    assert response.status_code == 429



import pytest

from threescale_api.resources import ApplicationPlan, Limits


@pytest.fixture()
def limit_client(application_plan, metric) -> Limits:
    return application_plan.limits(metric)


@pytest.fixture()
def limits(metric, application_plan: ApplicationPlan):
    params = dict(period='minute', value=10)
    application_plan.limits(metric).create(params)
    return application_plan.limits(metric).list()


def test_create_limit(limits):
    assert limits is not None
    limit = limits[0]
    assert limit['period'] == 'minute'
    assert limit['value'] == 10



import os
import secrets
from distutils.util import strtobool

import pytest
from dotenv import load_dotenv

import threescale_api
from threescale_api.resources import (Service, ApplicationPlan, Application,
                                      Proxy, Backend, Metric, MappingRule,
                                      BackendMappingRule, BackendUsage)

load_dotenv()

def cleanup(resource):
    resource.delete()
    assert not resource.exists()

def get_suffix() -> str:
    return secrets.token_urlsafe(8)


@pytest.fixture(scope='session')
def url() -> str:
    return os.getenv('THREESCALE_PROVIDER_URL')


@pytest.fixture(scope='session')
def token() -> str:
    return os.getenv('THREESCALE_PROVIDER_TOKEN')


@pytest.fixture(scope='session')
def master_url() -> str:
    return os.getenv('THREESCALE_MASTER_URL')


@pytest.fixture(scope='session')
def master_token() -> str:
    return os.getenv('THREESCALE_MASTER_TOKEN')


@pytest.fixture(scope="session")
def ssl_verify() -> bool:
    ssl_verify = os.getenv('THREESCALE_SSL_VERIFY', 'false')
    return bool(strtobool(ssl_verify))


@pytest.fixture(scope='session')
def api_backend() -> str:
    return os.getenv('TEST_API_BACKEND', 'http://www.httpbin.org:80')


@pytest.fixture(scope='session')
def api(url: str, token: str,
        ssl_verify: bool) -> threescale_api.ThreeScaleClient:
    return threescale_api.ThreeScaleClient(url=url, token=token,
                                           ssl_verify=ssl_verify)


@pytest.fixture(scope='session')
def master_api(master_url: str, master_token: str,
               ssl_verify: bool) -> threescale_api.ThreeScaleClient:
    return threescale_api.ThreeScaleClient(url=master_url, token=master_token,
                                           ssl_verify=ssl_verify)


@pytest.fixture(scope='module')
def apicast_http_client(application, ssl_verify):
    return application.api_client(verify=ssl_verify)


@pytest.fixture(scope='module')
def service_params():
    suffix = get_suffix()
    return dict(name=f"test-{suffix}")


@pytest.fixture(scope='module')
def service(service_params, api) -> Service:
    service = api.services.create(params=service_params)
    yield service
    cleanup(service)


@pytest.fixture(scope='module')
def account_params():
    suffix = get_suffix()
    name = f"test-{suffix}"
    return dict(name=name, username=name, org_name=name)


@pytest.fixture(scope='module')
def account(account_params, api):
    entity = api.accounts.create(params=account_params)
    yield entity
    cleanup(entity)


@pytest.fixture(scope='module')
def application_plan_params() -> dict:
    suffix = get_suffix()
    return dict(name=f"test-{suffix}")


@pytest.fixture(scope='module')
def application_plan(api, service, application_plan_params) -> ApplicationPlan:
    resource = service.app_plans.create(params=application_plan_params)
    yield resource


@pytest.fixture(scope='module')
def application_params(application_plan):
    suffix = get_suffix()
    name = f"test-{suffix}"
    return dict(name=name, description=name, plan_id=application_plan['id'])


@pytest.fixture(scope='module')
def application(account, application_plan, application_params) -> Application:
    resource = account.applications.create(params=application_params)
    yield resource
    cleanup(resource)


@pytest.fixture(scope='module')
def proxy(service, application, api_backend) -> Proxy:
    params = {
        'api_backend': api_backend,
        'credentials_location': 'query',
        'api_test_path': '/get',
    }
    proxy = service.proxy.update(params=params)
    return proxy


@pytest.fixture(scope='module')
def backend_usage(service, backend, application) -> BackendUsage:
    params = {
        'service_id': service['id'],
        'backend_api_id': backend['id'],
        'path': '/get',
    }
    resource = service.backend_usages.create(params=params)
    yield resource
    cleanup(resource)

@pytest.fixture(scope='module')
def metric_params(service):
    suffix = get_suffix()
    friendly_name = f'test-metric-{suffix}'
    system_name = f'{friendly_name}'.replace('-', '_')
    return dict(service_id=service['id'], friendly_name=friendly_name,
                system_name=system_name, unit='count')

@pytest.fixture(scope='module')
def backend_metric_params(backend):
    suffix = get_suffix()
    friendly_name = f'test-metric-{suffix}'
    system_name = f'{friendly_name}'.replace('-', '_')
    return dict(backend_id=backend['id'], friendly_name=friendly_name,
                system_name=system_name, unit='count')

@pytest.fixture
def updated_metric_params(metric_params):
    suffix = get_suffix()
    friendly_name = f'test-updated-metric-{suffix}'
    metric_params['friendly_name'] = f'/anything/{friendly_name}'
    metric_params['system_name'] = friendly_name.replace('-', '_')
    return metric_params

@pytest.fixture
def backend_updated_metric_params(backend_metric_params):
    suffix = get_suffix()
    friendly_name = f'test-updated-metric-{suffix}'
    backend_metric_params['friendly_name'] = f'/anything/{friendly_name}'
    backend_metric_params['system_name'] = friendly_name.replace('-', '_')
    return backend_metric_params



@pytest.fixture(scope='module')
def metric(service, metric_params) -> Metric:
    resource = service.metrics.create(params=metric_params)
    yield resource
    cleanup(resource)


@pytest.fixture(scope='module')
def hits_metric(service):
    return service.metrics.read_by(system_name='hits')


@pytest.fixture(scope='module')
def method_params(service):
    suffix = get_suffix()
    friendly_name = f'test-method-{suffix}'
    system_name = f'{friendly_name}'.replace('-', '_')
    return dict(friendly_name=friendly_name, system_name=system_name,
                unit='hits')


@pytest.fixture
def updated_method_params(method_params):
    suffix = get_suffix()
    friendly_name = f'test-updated-method-{suffix}'
    method_params['friendly_name'] = friendly_name
    method_params['system_name'] = f'{friendly_name}'.replace('-', '_')
    return method_params


@pytest.fixture(scope='module')
def method(hits_metric, method_params):
    resource = hits_metric.methods.create(params=method_params)
    yield resource
    cleanup(resource)


def get_mapping_rule_pattern():
    suffix = get_suffix()
    pattern = f'test-{suffix}'.replace('_', '-')
    return pattern


@pytest.fixture(scope='module')
def mapping_rule_params(hits_metric):
    """
    Fixture for getting paramteres for mapping rule for product/service.
    """
    return dict(http_method='GET', pattern='/', metric_id=hits_metric['id'],
                delta=1)


@pytest.fixture(scope='module')
def backend_mapping_rule_params(backend_metric):
    """
    Fixture for getting paramteres for mapping rule for backend.
    """
    return dict(http_method='GET', pattern='/get/anything/id', metric_id=backend_metric['id'],
                delta=1)

@pytest.fixture
def updated_mapping_rules_params(mapping_rule_params):
    """
    Fixture for updating mapping rule for product/service.
    """
    pattern = get_mapping_rule_pattern()
    params = mapping_rule_params.copy()
    params['pattern'] = f'/get/anything/{pattern}'
    return params

@pytest.fixture
def updated_backend_mapping_rules_params(backend_mapping_rule_params):
    """
    Fixture for updating mapping rule for backend.
    """
    pattern = get_mapping_rule_pattern()
    params = backend_mapping_rule_params.copy()
    params['pattern'] = f'/get/anything/{pattern}'
    return params


@pytest.fixture(scope='module')
def mapping_rule(proxy, mapping_rule_params) -> MappingRule:
    """
    Fixture for getting mapping rule for product/service.
    """
    resource = proxy.mapping_rules.create(params=mapping_rule_params)
    yield resource
    cleanup(resource)

@pytest.fixture(scope='module')
def backend_mapping_rule(backend, backend_mapping_rule_params) -> BackendMappingRule:
    """
    Fixture for getting mapping rule for backend.
    """
    resource = backend.mapping_rules.create(params=backend_mapping_rule_params)
    yield resource
    cleanup(resource)

@pytest.fixture
def create_mapping_rule(service):
    """
    Fixture for creating mapping rule for product/service.
    """
    rules = []
    proxy = service.proxy.list()

    def _create(metric, http_method, path):
        params = dict(service_id=service['id'],
                      http_method=http_method,
                      pattern=f'/anything{path}',
                      delta=1, metric_id=metric['id'])
        rule = proxy.mapping_rules.create(params=params)
        rules.append(rule)
        return rule

    yield _create

    for rule in rules:
        if rule.exists():
            cleanup(rule)

@pytest.fixture
def create_backend_mapping_rule(backend):
    """
    Fixture for creating mapping rule for backend.
    """
    rules = []

    def _create(backend_metric, http_method, path):
        params = dict(backend_id=backend['id'],
                      http_method=http_method,
                      pattern=f'/anything{path}',
                      delta=1, metric_id=backend_metric['id'])
        rule = backend.mapping_rules.create(params=params)
        rules.append(rule)
        return rule

    yield _create

    for rule in rules:
        if rule.exists():
            cleanup(rule)

@pytest.fixture(scope='module')
def backend_params(api_backend):
    """
    Fixture for getting backend parameters.
    """
    suffix = get_suffix()
    return dict(name=f"test-backend-{suffix}",
                private_endpoint=api_backend,
                description='111')

@pytest.fixture(scope='module')
def backend(backend_params, api) -> Backend:
    """
    Fixture for getting backend.
    """
    backend = api.backends.create(params=backend_params)
    yield backend
    cleanup(backend)

@pytest.fixture(scope='module')
def backend_metric(backend, metric_params) -> Metric:
    """
    Fixture for getting backend metric.
    """
    resource = backend.metrics.create(params=metric_params)
    yield resource
    cleanup(resource)



import pytest


def create_resources(service, proxy, metric_params):
    metric = service.metrics.create(params=metric_params)
    rule_params = dict(http_method='GET', metric_id=metric['id'], delta=1,
                       pattern=f'/anything/{metric["system_name"]}')
    rule = proxy.mapping_rules.create(params=rule_params)
    return rule['pattern']


def do_request(client, path):
    for _ in range(2):
        response = client.get(path=path)
        assert response.status_code == 200


def test_should_get_analytics_by_service(api, service, proxy, metric_params,
                                         apicast_http_client):
    path = create_resources(service, proxy, metric_params)
    do_request(apicast_http_client, path)
    data = api.analytics.list_by_service(service)
    assert data['total'] == 2

